#define _USE_MATH_DEFINES
#include <cmath>

#include "Hemicube.h"

Hemicube::Hemicube(const unsigned int& resolution, const GLpoint3f& centre, const GLpoint3f& normal, Scene* scene) : resolution(resolution), centre(centre), scene(scene)
{
   top = 0;
   bottom = 0;
   left = 0;
   right = 0;
   front = 0;

   zAxis = normal;
/*
If zAxis is too near to the vector (1, 0, 0) use (0, 1, 0) to generate perpendicular vectors.
Note that 1/sqrt(2) gives perfect balancing for when when we choose one case or the other.
*/
   if(fabs(zAxis.x) > M_SQRT1_2)
   {
      xAxis = zAxis.cross(GLpoint3f(0, 1, 0));
   }
   else
   {
      xAxis = zAxis.cross(GLpoint3f(1, 0, 0));
   }
   yAxis = zAxis.cross(xAxis);
}

Hemicube::~Hemicube(void)
{
   delete [] top;
   delete [] bottom;
   delete [] left;
   delete [] right;
   delete [] front;
}

void Hemicube::rotate(const GLfloat& angle)
{
   GLpoint3f newX, newY;
   newX = xAxis * cos(angle) - yAxis * sin(angle);
   newY = xAxis * sin(angle) + yAxis * cos(angle);
   xAxis = newX;
   yAxis = newY;
   return;
}

void Hemicube::render_face(const GLpoint3f& xAxis, const GLpoint3f& yAxis, const GLpoint3f& zAxis, GLubyte *image)
{
   glShadeModel(GL_FLAT);
   glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

   GLfloat *model_view;
   model_view = new GLfloat[16];
   // NB: The below is correct and not incorrect by a transpose because glLoadMatrix has an unusual convention.
   model_view[0] = xAxis.x;    model_view[1] = yAxis.x;    model_view[2] = zAxis.x;    model_view[3] = 0;
   model_view[4] = xAxis.y;    model_view[5] = yAxis.y;    model_view[6] = zAxis.y;    model_view[7] = 0;
   model_view[8] = xAxis.z;    model_view[9] = yAxis.z;    model_view[10] = zAxis.z;   model_view[11] = 0;
   model_view[12] = 0;         model_view[13] = 0;         model_view[14] = 0;         model_view[15] = 1;

   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glLoadMatrixf(model_view);
   delete [] model_view;
   glTranslatef(-centre.x, -centre.y, -centre.z);

   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
   glLoadIdentity();
   /*
      The arguments to glFrustum below should really be parameters in the hemicube instead of literals,
      particularly the last one. 
   */
   glFrustum(-1.0f, 1.0f, -1.0f, 1.0f, 1.0f, 1000.0f);
   
   glPushAttrib(GL_VIEWPORT_BIT);
   glViewport(0, 0, resolution, resolution);

   unsigned int i, j, k, patch_ind = 1; // Starting patch_ind at 1 means no patch has colour 0 so we can detect the background if we need to.
   GLpoint3f vertex;
   for(j = 0; j < scene->n_meshes; j++)
   {
      for(i = 0; i < scene->meshes[j].n_patches; i++)
      {
         glBegin(GL_TRIANGLES);
         glColor3ubv((GLubyte *) &patch_ind);
         for(k = 0; k < 3; k++)
         {
            vertex = scene->meshes[j].vertices[scene->meshes[j].patches[i].vertex_ind[k]];
            glVertex3f(vertex.x, vertex.y, vertex.z);
         }
         glEnd();
         patch_ind++;
      }
   }
   glReadBuffer(GL_BACK); // Possibly unnecessary.
   glReadPixels(0, 0, resolution, resolution, GL_RGB, GL_UNSIGNED_BYTE, image); // Ouch!

   glMatrixMode(GL_MODELVIEW);
   glPopMatrix();
   glMatrixMode(GL_PROJECTION);
   glPopMatrix();
   glPopAttrib();
   glShadeModel(GL_SMOOTH); // Theoretically, I should check this was the case before we set flat shading above.
   return;
}

void Hemicube::render_all(void)
{
   unsigned int alloc_size = 3 * resolution * resolution; // The 3 is because I'm indexing patches using RGB in render_face(...). I may switch to RGBA and in that case, this 3 better become a 4.

   front = new GLubyte[alloc_size];
   render_face(xAxis, yAxis, zAxis, front);

   /*
   Of course, we only ever use half data rendered into the 4 buffers that represent
   sides of the cube (ie: the sides of the hemicube). We thus need only render into
   the half sides and this can be accomplished by using an appropriate clipping
   plane. I will impliment this later (though it won't make much, if any, difference
   to the speed). Alternatively, I suppose I could also just use a glScissor. I'll
   decide later which is better.
   */
   top = new GLubyte[alloc_size];
   render_face(xAxis, zAxis, yAxis * -1.0f, top);
   //render_face(xAxis * -1.0f, zAxis * -1.0f, yAxis * -1.0f, top);   

   bottom = new GLubyte[alloc_size];
   //render_face(xAxis, zAxis * -1.0f, yAxis, bottom);
   render_face(xAxis * -1.0f, zAxis, yAxis, bottom);

   right = new GLubyte[alloc_size];
   render_face(zAxis, yAxis, xAxis * -1.0f, right);
   //render_face(zAxis * -1.0f, yAxis * -1.0f, xAxis * -1.0f, right);   

   left = new GLubyte[alloc_size];
   //render_face(zAxis * -1.0f, yAxis, xAxis, left);
   render_face(zAxis, yAxis * -1.0f, xAxis, left);
   
   return;
}