#include "Mesh.h"
#include "HashTable.h"
#include "misc.h"

Mesh::Mesh(void)
{
   n_vertices = 0;
   n_patches = 0;
}

Mesh::~Mesh(void)
{
   delete [] vertex_radiosity;
   delete [] patch_ind;
}

void Mesh::setFace(int index, int a, int b, int c)
{
   if(index >= max_patches) return;
   patches[index].vertex_ind[0] = a;
   patches[index].vertex_ind[1] = b;
   patches[index].vertex_ind[2] = c;
   return;
}

void Mesh::setVertex(int index, float x, float y, float z)
{
   if(index >= max_vertices) return;
   vertices[index] = GLpoint3f(x, y, z);
   return;
}

int Mesh::getVertCnt(void)
{
   return n_vertices;
}

int Mesh::getFaceCnt(void)
{
   return n_patches;
}

void Mesh::subdivide(void)
{
   /*
   This method replaces every triangle in the scene with four new triangles.
   It does this by adding vertices at the midpoints of all edges and then
   for a given triangle forming a new triangle by joining up these midpoints to
   give a triangle. The complement of this new triangle in the old one has
   three connected components and these are the other three new triangles.
   I think this may be known as Loop subdivision.
   */
   HashTable new_vertices(next_prime(max_vertices)); // Dunno what the ideal size is here but probably best if it's prime.

   unsigned int i, j, new_vertex_ind[3], a, b, N = n_patches;

   for(i = 0; i < N; i++)
   {
      if(n_vertices + 3 > max_vertices || n_patches + 3 > max_patches)
      {
         break;
      }

      // First create the new vertices if necessary.
      for(j = 0; j < 3; j++)
      {
         a = patches[i].vertex_ind[(j + 1) % 3];
         b = patches[i].vertex_ind[(j + 2) % 3];
         if(a > b) std::swap(a, b);
         if(!new_vertices.look_up(a, b, new_vertex_ind[j]))
         {
            vertices[n_vertices] = (vertices[a] + vertices[b]) * 0.5f;
            new_vertices.add(a, b, n_vertices);
            new_vertex_ind[j] = n_vertices++;
         }
      }

      // OK, now we've got indices for the new vertices. Let's make the patches.
      patches[i].area *= 0.25; // Actually, the way I use this method the areas aren't usually specified at this stage anyway.
      for(j = 0; j < 3; j++)
      {
         patches[n_patches] = patches[i];
         patches[n_patches].vertex_ind[0] = patches[i].vertex_ind[j];
         patches[n_patches].vertex_ind[1] = new_vertex_ind[(j + 2) % 3]; // Note that I'm careful about the ordering of vertices.
         patches[n_patches].vertex_ind[2] = new_vertex_ind[(j + 1) % 3];
         n_patches++;
      }
      for(j = 0; j < 3; j++)
      {
         patches[i].vertex_ind[j] = new_vertex_ind[j];
      }
   }
   return;
}

void Mesh::initialize(const unsigned int n_subdivisions, const GLRGBColour& colour, const GLfloat& reflectance, const GLRGBColour& emission)
{
   unsigned int i;
   for(i = 0; i < n_subdivisions; i++)
   {
      subdivide();
   }

   for(i = 0; i < n_patches; i++)
   {
      patches[i].area = get_area(vertices[patches[i].vertex_ind[0]], vertices[patches[i].vertex_ind[1]], vertices[patches[i].vertex_ind[2]]);
      patches[i].radiosity = emission;
      patches[i].unsent_radiosity = emission;
   }

   this->reflectance = reflectance;
   this->colour = colour;
   vertex_radiosity = new GLRGBColour[n_vertices];
   patch_ind = new std::list<unsigned int>[n_vertices];
   update_patch_ind();
   update_vertex_radiosities();
   return;
}

void Mesh::update_patch_ind(void)
{
   unsigned int i, j;

   for(i = 0; i < n_patches; i++)
   {
      for(j = 0; j < 3; j++)
      {
         patch_ind[patches[i].vertex_ind[j]].push_front(i);
      }
   }
   return;
}

void Mesh::update_vertex_radiosities(void)
{
   unsigned int i;
   std::list<unsigned int>::iterator it;

   for(i = 0; i < n_vertices; i++)
   {
      vertex_radiosity[i].set_black();
      for(it = patch_ind[i].begin(); it != patch_ind[i].end(); it++)
      {
         vertex_radiosity[i] += patches[*it].radiosity; // Perhaps I should add area weighting to this.
      }
      vertex_radiosity[i] /= patch_ind[i].size();
   }
  return;
}