#define _USE_MATH_DEFINES
#include <cmath>
#include <ctime>
#include <iostream>

#include "Scene.h"
#include "Hemicube.h"
#include "misc.h"
#include "uint24.h"
#include "3DS Loader/FileSys.h"

Scene::Scene(void)
{   
   unsigned int i, min_subdivisions = 2;
   GLfloat brightness = 15.0f, reflectance = 1.0f;

   n_meshes = 0;

   GLRGBColour floor_colour(1.0f, 0.3f, 0.3f), column_colour(0.7f, 1.0f, 0.7f), wall_colour(1.0f, 1.0f, 1.0f);
   GLRGBColour alcove_colour(1.0f, 1.0f, 0.3f), ceiling_colour(1.0f, 1.0f, 1.0f);

   meshes = new Mesh[21];
   FileSys::load3ds("C:\\Documents and Settings\\Oliver\\My Documents\\Visual Studio 2005\\Projects\\progressive radiosity\\models\\simproom.3ds", -1, this);
   
   meshes[0].initialize(min_subdivisions, wall_colour, reflectance, Black);
   meshes[1].initialize(min_subdivisions + 4, floor_colour, reflectance, Black);
   meshes[2].initialize(min_subdivisions + 4, ceiling_colour, reflectance, Black);
   for(i = 3; i <= 5; i++)
   {
      meshes[i].initialize(min_subdivisions + 2, wall_colour, reflectance, Black);
   }
   for(i = 6; i <= 13; i++)
   {
      meshes[i].initialize(min_subdivisions + 2, column_colour, reflectance, Black);
   }
   meshes[14].initialize(min_subdivisions , White, reflectance, White * brightness);
   meshes[15].initialize(min_subdivisions, White, reflectance, White * brightness);
   for(i = 16; i <= 20; i++)
   {
      meshes[i].initialize(min_subdivisions + 2, alcove_colour, reflectance, Black);
   }

   hemicube_pixel_size = 2.0f / hemicube_resolution; // This 2.0f is determined by the literals in my call to glFrustum in the render_face method of the hemicube class. I'll get rid of the literals later.
   n_patches = 0;
   for(i = 0; i < n_meshes; i++)
   {
      n_patches += meshes[i].n_patches;
      std::cout << "mesh[" << i << "] has " << meshes[i].n_patches << " patches and " << meshes[i].n_vertices << " vertices." << std::endl;
   }
}

Scene::~Scene()
{
   delete [] meshes;
}

Mesh *Scene::addMesh(char *id, int vertCnt, int faceCnt)
{
   meshes[n_meshes].n_vertices = vertCnt;
   meshes[n_meshes].n_patches = faceCnt;
   return &(meshes[n_meshes++]);
}

void Scene::get_form_factors(const Hemicube& cube, GLfloat *form_factors)
{
   unsigned int i;
   for(i = 0; i < n_patches + 1; i++) // See the relevant line in Hemicube.render_face(...) to see why we add 1 here.
   {
      form_factors[i] = 0.0f;
   }

   unsigned int u, v;
   uint24 *(hemicube_faces[5]);
   hemicube_faces[0] = (uint24 *) cube.front;
   hemicube_faces[1] = (uint24 *) cube.top;
   hemicube_faces[2] = (uint24 *) cube.bottom;
   hemicube_faces[3] = (uint24 *) cube.right;
   hemicube_faces[4] = (uint24 *) cube.left;

   GLfloat x, y, ff_inc;
   for(u = 0; u < hemicube_resolution; u++)
   {
      for(v = 0; v < hemicube_resolution; v++)
      {
         x = u * hemicube_pixel_size - 1.0f;
         y = v * hemicube_pixel_size - 1.0f;
         // Front face of hemicube.
         form_factors[hemicube_faces[0][u + v * hemicube_resolution].x()] += 1.0f / sqr(sqr(x) + sqr(y) + 1.0f);
         if(v < hemicube_resolution / 2)
         {
            ff_inc = -y / sqr(sqr(x) + sqr(y) + 1.0f);
            form_factors[hemicube_faces[1][u + v * hemicube_resolution].x()] += ff_inc; // Top face.
            form_factors[hemicube_faces[2][u + v * hemicube_resolution].x()] += ff_inc; // Bottom face.
         }
         if(u < hemicube_resolution / 2)
         {
            ff_inc = -x / sqr(sqr(x) + sqr(y) + 1.0f);
            form_factors[hemicube_faces[3][u + v * hemicube_resolution].x()] += ff_inc; // Right face.
            form_factors[hemicube_faces[4][u + v * hemicube_resolution].x()] += ff_inc; // Left face.
         }
      }
   }
   for(i = 0; i < 1 + n_patches; i++)
   {
      form_factors[i] *= sqr(hemicube_pixel_size) / M_PI;
      /*if(form_factors[i] >= 1.0f)
      {
         form_factors[i] = 1.0f;
      }*/
   }
   return;
}

void Scene::run_pass(void)
{
   unsigned int i, j, max_unsent_mesh_ind, max_unsent_patch_ind;
   GLfloat max_unsent_energy = 0.0f, energy;

   for(j = 0; j < n_meshes; j++)
   {
      for(i = 0; i < meshes[j].n_patches; i++)
      {
         energy = (meshes[j].patches[i].unsent_radiosity.r + meshes[j].patches[i].unsent_radiosity.g + meshes[j].patches[i].unsent_radiosity.b) * meshes[j].patches[i].area;
         if(energy > max_unsent_energy)
         {
            max_unsent_mesh_ind = j;
            max_unsent_patch_ind = i;
            max_unsent_energy = energy;
         }
      }
   }
   unsigned int *vertex_ind = meshes[max_unsent_mesh_ind].patches[max_unsent_patch_ind].vertex_ind;
   GLpoint3f v1 = meshes[max_unsent_mesh_ind].vertices[vertex_ind[0]];
   GLpoint3f v2 = meshes[max_unsent_mesh_ind].vertices[vertex_ind[1]];
   GLpoint3f v3 = meshes[max_unsent_mesh_ind].vertices[vertex_ind[2]];
   
   /*
   Note that we have to be careful below to decide whether we need normal or -normal. In
   order to be sure, we need to have a consistent ordering of the vertices on the patches.
   */
   GLpoint3f normal = (v3 - v1).cross(v2 - v1);
   normal.normalise();
   Hemicube cube(hemicube_resolution, (v1 + v2 + v3) * 0.333333f, normal, this);
   cube.rotate(M_PI * rand() / 16384); // Significantly reduces an aliasing problem.
   cube.render_all();
   
/*   saveBMP24("C:\\Documents and Settings\\Oliver\\Desktop\\hemicube_front.bmp", cube.front, cube.resolution, cube.resolution);
   saveBMP24("C:\\Documents and Settings\\Oliver\\Desktop\\hemicube_top.bmp", cube.top, cube.resolution, cube.resolution);
   saveBMP24("C:\\Documents and Settings\\Oliver\\Desktop\\hemicube_bottom.bmp", cube.bottom, cube.resolution, cube.resolution);
   saveBMP24("C:\\Documents and Settings\\Oliver\\Desktop\\hemicube_left.bmp", cube.left, cube.resolution, cube.resolution);
   saveBMP24("C:\\Documents and Settings\\Oliver\\Desktop\\hemicube_right.bmp", cube.right, cube.resolution, cube.resolution);
*/
   GLfloat *form_factors = new GLfloat[1 + n_patches]; // See the relevant line in Hemicube.render_face(...) to see why we add 1 here.
   get_form_factors(cube, form_factors);

   unsigned int patch_ind = 1; // See the relevant line in Hemicube.render_face(...) to see why we start this at 1.
   // The loop below is the progressive radiosity algorithm. Simple eh!
   GLRGBColour radiosity_change; 
   GLRGBColour unsent_radiosity = meshes[max_unsent_mesh_ind].patches[max_unsent_patch_ind].unsent_radiosity;
   GLfloat area = meshes[max_unsent_mesh_ind].patches[max_unsent_patch_ind].area;
   for(j = 0; j < n_meshes; j++)
   {
      if(j == max_unsent_mesh_ind) // Assuming meshes are always planar (or at least not concave) we can skip this value.
      {
         patch_ind += meshes[j].n_patches;
         continue;
      }
      for(i = 0; i < meshes[j].n_patches; i++)
      {
         radiosity_change = meshes[j].colour * unsent_radiosity * (meshes[j].reflectance *
                            form_factors[patch_ind++] * area / meshes[j].patches[i].area); 
         meshes[j].patches[i].unsent_radiosity += radiosity_change;
         meshes[j].patches[i].radiosity += radiosity_change;
      }
   }
   meshes[max_unsent_mesh_ind].patches[max_unsent_patch_ind].unsent_radiosity.set_black();
   delete [] form_factors;
   return;
}