#define _USE_MATH_DEFINES
#include <cmath>
#include <list>
#include <iostream>
#include <windows.h>
#include <GL/glut.h>

#include "Scene.h"
#include "misc.h"

const float Z_NEAR_CLIP         = 1.0f;
const float Z_FAR_CLIP          = 1000.0f;
const float FOV                 = M_PI_2;
const int   WINDOW_SIZE         = 800;
const float TONE_MAPPING_FACTOR = 1.0f;

void mouseMotionFunc(int x, int y);
void keyboardFunc(unsigned char key, int x, int y);
void windowResizeFunc(GLsizei width, GLsizei height);

void initGL();
void renderScene();
void updateScene();

Scene *my_scene;
int x_angle, y_angle;
GLpoint3f position(0.0f, 0.0f, 0.0f);
bool flat_shade = false;

int main(int argc, char** argv)
{
   my_scene = new Scene();
   x_angle = 0;
   y_angle = 0;

   glutInit(&argc, argv);
   
   glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
   glutInitWindowPosition(50, 50);
   glutInitWindowSize(WINDOW_SIZE, WINDOW_SIZE);

   int id = glutCreateWindow("Progressive radiosity.");

   glutKeyboardFunc(keyboardFunc);
   glutMotionFunc(mouseMotionFunc);
   glutReshapeFunc(windowResizeFunc);
   glutDisplayFunc(renderScene);
   glutIdleFunc(updateScene);

   initGL();

   glutMainLoop();
   glutDestroyWindow(id);
   delete my_scene;
   return 0;
}

void windowResizeFunc(GLsizei width, GLsizei height)
{
   glViewport(0, 0, width, height);
   return;
}

void initGL()
{
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();							
   float aspect = 1.0f; // Because we've only got WINDOW_SIZE.
   float h = (float) tan(FOV / 2) * Z_NEAR_CLIP;
   float w = aspect * h;
   glFrustum(-w, w, -h, h, Z_NEAR_CLIP, Z_FAR_CLIP);
   
   glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
   glClearDepth(1.0f);
   glDepthFunc(GL_LESS);
   glEnable(GL_DEPTH_TEST);
   glDisable(GL_LIGHTING);
   glShadeModel(GL_SMOOTH);
   /* glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
   glEnable(GL_TEXTURE_2D);
   glEnable(GL_CULL_FACE); */
   return;
}

void mouseMotionFunc(int x, int y)
{
   return;
}

void keyboardFunc(unsigned char key, int x, int y)
{
   static const GLfloat speed = 2.0f, angle_speed = 3.0f;
   static GLfloat position_x_inc = speed, position_z_inc = 0;
   unsigned int i;
   
   switch(key)
   {
   case 'p':
      my_scene->run_pass();
      for(i = 0; i < my_scene->n_meshes; i++)
      {
         my_scene->meshes[i].update_vertex_radiosities();
      }
      break;
   case 'o':
      for(i = 0; i < 50; i++)
      {
         my_scene->run_pass();
      }
      for(i = 0; i < my_scene->n_meshes; i++)
      {
         my_scene->meshes[i].update_vertex_radiosities();
      }
      break;
   case 'd':
      position.x += position_x_inc;
      position.z += position_z_inc;
      break;
   case 'a':
      position.x -= position_x_inc;
      position.z -= position_z_inc;
      break;
   case 'w':
      position.x += position_z_inc;
      position.z += -position_x_inc;
      break;
   case 's':
      position.x -= position_z_inc;
      position.z -= -position_x_inc;
      break;
   case 'z':
      position.y += speed;
      break;
   case 'c':
      position.y -= speed;
      break;
   case 'j':
      y_angle += angle_speed;
      position_x_inc = speed * cos(y_angle / 180.0f * M_PI);
      position_z_inc = -speed * sin(y_angle / 180.0f * M_PI);
      break;
   case 'l':
      y_angle -= angle_speed;
      position_x_inc = speed * cos(y_angle / 180.0f * M_PI);
      position_z_inc = -speed * sin(y_angle / 180.0f * M_PI);
      break;
   case 'k':
      x_angle += angle_speed;
      break;
   case 'i':
      x_angle -= angle_speed;
      break;
   case 't':
      if(flat_shade == true)
      {
         flat_shade = false;
      }
      else
      {
         flat_shade = true;
      }
      break;
   }
   return;
}

void renderScene()
{
   unsigned int i, j, k, n;
   GLRGBColour colour;
   GLpoint3f vertex;

   glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glRotatef(-x_angle, 1.0f, 0.0f, 0.0f);
   glRotatef(-y_angle, 0.0f, 1.0f, 0.0f);
   glTranslatef(-position.x, -position.y, -position.z);

   for(k = 0; k < my_scene->n_meshes; k++)
   {
      for(i = 0; i < my_scene->meshes[k].n_patches; i++)
      {
         glBegin(GL_TRIANGLES);
         for(j = 0; j < 3; j++)
         {
            n = my_scene->meshes[k].patches[i].vertex_ind[j];
            vertex = my_scene->meshes[k].vertices[n];
            if(flat_shade == false)
            {
               colour = my_scene->meshes[k].vertex_radiosity[n];
            }
            else
            {
               colour = my_scene->meshes[k].patches[i].radiosity;
            }
            colour.tone_map(TONE_MAPPING_FACTOR);
            glColor3f(colour.r, colour.g, colour.b);
            glVertex3f(vertex.x, vertex.y, vertex.z);
         }
         glEnd();
      }
   }

   glutSwapBuffers();
   return;
}

void updateScene()
{
   static DWORD lastTickCount = timeGetTime();
   while(timeGetTime() - lastTickCount < 32) ;
   lastTickCount = timeGetTime();

   glutPostRedisplay();
   return;
}